//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#include <cerrno>

#include <lib/debug.h>
#include <lib/directory.h>
#include <lib/hexdump.h>
#include <lib/quit.h>
#include <lib/sector_io/apple.h>
#include <lib/sector_io/mmap.h>
#include <lib/sector_io/offset.h>
#include <lib/sector_io/pdp.h>
#include <lib/sector_io/raw.h>


static bool
has_valid_signature(sector_io_smart_pointer &sio)
{
    unsigned char data[256];
    if (sio->read(1024, data, 256) < 0)
	return false;
    DEBUG(4, "%s", hexdump(data, 16).c_str());
    return
	(
	    // dfirstblock == 0
	    data[0] == 0
	&&
	    data[1] == 0
	&&
	    // dlastblock == 6
	    (
		(data[2] == 0 && (data[3] == 6 || data[3] == 10))
	    ||
		((data[2] == 6 || data[2] == 10) && data[3] == 0)
	    )
	&&
	    // volume name length is valid
	    data[6] > 0
	&&
	    data[6] < 8
	);
}

/**
  * The interleaved_raw_sector_io function is used to figure out what
  * interleaving (if any) is needed for the disk image.
  *
  * We have to sniff the disk image to see if we need to interleave
  * it or not.  This is all very heuristic.  I have only got Apple ][
  * Pascal images to work with.  Some are interleaved, and some are
  * not.
  *
  * Patches to extend this code to understand more disk image formats
  * are most welcome.
  *
  * @param filename
  *     The name of the file containing the disk image.
  * @returns
  *     pointer to sector i/o for accessing disk image
  */
static sector_io_smart_pointer
interleaved_raw_sector_io(const rcstring &filename, bool read_only)
{

    //
    // Open the file.  Try for memory mapped I/O, and fall back to
    // regular file I/O if not available.
    //
    sector_io_smart_pointer raw;
    if (sector_io_mmap::available(filename, read_only))
	raw = new sector_io_mmap(filename, read_only);
    else
	raw = new sector_io_raw(filename, read_only);

    //
    // See if the raw disk image has a valid signature.
    // This is the ideal case.
    if (has_valid_signature(raw))
	return raw;

    //
    // See if it has Apple ][ Pascal sector interleaving.
    //
    sector_io_smart_pointer trial = new sector_io_apple(raw);
    if (has_valid_signature(trial))
	return trial;

    //
    // See if it has PDP11 interleaving.  There are three cases: offset,
    // offset and interleaved and interleaved by itself.  (There are
    // actually four cases, but we already tested the ideal case.)
    //
    // The PDP11 disks have 77 tracks of 26 sectors of 128 bytes, but
    // the first track is ignored.
    //
    trial = new sector_io_offset(raw, 128 * 26);
    if (has_valid_signature(trial))
	return trial;

    sector_io_smart_pointer trial2 = new sector_io_pdp(trial);
    if (has_valid_signature(trial2))
	return trial2;

    trial = new sector_io_pdp(raw);
    if (has_valid_signature(trial))
	return trial;

    //
    // No luck.
    //
    quitter.fatal_error("%s: unable to find volume label", filename.c_str());
    return raw;
}


directory *
directory::factory(const rcstring &filename, bool read_only, concern_t level)
{
    if (read_only && level > concern_check)
	level = concern_check;
    sector_io_smart_pointer disk =
	interleaved_raw_sector_io(filename, read_only);
    directory *dir = new directory(disk);
    int err = dir->meta_read(level);
    if (err < 0)
    {
	//
	// If we can't read the volume, make sure the mount fails as well.
	//
	errno = -err;
	quitter.fatal_error_with_errno("read %s", filename.c_str());
    }
    if (err > 0)
    {
	if (level >= concern_repair)
	{
	    quitter.fatal_error
    	    (
		"%s: repaired %d errors",
		filename.c_str(),
		err
	    );
	    // NOTREACHED
	}
	quitter.message
	(
	    "%s: warning: found %d format errors",
	    filename.c_str(),
	    err
	);
    }
    return dir;
}
