//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#ifndef LIB_SECTOR_IO_SMART_POINTER_H
#define LIB_SECTOR_IO_SMART_POINTER_H

#include <lib/sector_io.h>

/**
  * The sector_io_smart_pointer class is used to represent a "smart
  * pointer", which keeps track of reference counts, and deletes the
  * objects pointed to when all references are gone.
  */
class sector_io_smart_pointer
{
public:
    /**
      * The destructor.
      */
    virtual ~sector_io_smart_pointer();

    /**
      * The constructor.
      *
      * @param a_ref
      *     The "dumb" pointer to be managed.  The reference count will
      *     not be increased, it is assumed you are giving thre dumb
      *     pointer to this smart pointer to manage it.
      */
    sector_io_smart_pointer(sector_io *a_ref) : ref(a_ref) { }

    /**
      * The default constructor.
      */
    sector_io_smart_pointer();

    /**
      * The copy constructor.
      */
    sector_io_smart_pointer(const sector_io_smart_pointer &);

    /**
      * The assignment operator.
      */
    sector_io_smart_pointer &operator=(const sector_io_smart_pointer &);

    /**
      * The "dereference member of" operator is used to obtain the
      * object to actually dereference, rather than this object.  This
      * is why mkes it a "smart pointer" when it actually isn't either.
      *
      * @note
      *     You will get segfaults if you try to dereference a NULL
      *     pointer, just like with a real pointer.
      */
    sector_io *operator->() { return ref; }

    /**
      * The "dereference member of" operator is used to obtain the
      * object to actually dereference, rather than this object.  This
      * is why mkes it a "smart pointer" when it actually isn't either.
      *
      * @note
      *     You will get segfaults if you try to dereference a NULL
      *     pointer, just like with a real pointer.
      */
    const sector_io *operator->() const { return ref; }

    /**
      * The close method is used to cancel this refernce to the
      * sectio_io object.
      */
    void close();

private:
    /**
      * The ref instance variable is used to remember the sector_io
      * object instance being managed.  It is assumed that the object
      * being managed is on the heap (as opposed to on the stack, or a
      * global variable).
      */
    sector_io *ref;
};

#endif // LIB_SECTOR_IO_SMART_POINTER_H
