//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <fcntl.h>
#include <unistd.h>

#include <lib/debug.h>
#include <lib/directory.h>
#include <lib/directory/entry.h>
#include <lib/progname.h>
#include <lib/quit.h>
#include <lib/version.h>


static void
usage()
{
    const char *prog = progname_get();
    fprintf(stderr, "Usage: %s -f <disk.image> -l\n", prog);
    fprintf(stderr, "       %s -f <disk.image> -g <file.to.get>...\n", prog);
    fprintf(stderr, "       %s -f <disk.image> -p <file.to.put>...\n", prog);
    fprintf(stderr, "       %s -f <disk.image> -r <file.to.remove>...\n", prog);
    fprintf(stderr, "       %s -V\n", prog);
    quitter.exit(1);
}


static mode_t
get_umask()
{
    mode_t um = umask(0);
    umask(um);
    return (um & 0777);
}


static void
get(directory *volume, const char *filename)
{
    directory_entry *dep = volume->find(filename);
    if (!dep)
    {
	errno = ENOENT;
	quitter.fatal_error_with_errno
	(
	    "open %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }

    int fd = open(filename, O_WRONLY | O_TRUNC | O_CREAT, 0666);
    if (fd < 0)
	quitter.fatal_error_with_errno("open %s", filename);

    int err = dep->open();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "open %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }

    off_t offset = 0;
    for (;;)
    {
	char buffer[1 << 14];
	int n = dep->read(offset, buffer, sizeof(buffer));
	if (n < 0)
	{
	    errno = -n;
	    quitter.fatal_error_with_errno
    	    (
		"read %s:%s",
		volume->get_volume_name().c_str(),
		filename
	    );
	}
	if (n == 0)
	    break;
	err = write(fd, buffer, n);
	if (err < 0)
	    quitter.fatal_error_with_errno("write %s", filename);
	offset += n;
    }
    err = dep->flush();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "flush %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
    err = dep->release();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "release %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
    close(fd);
}


static void
put(directory *volume, const char *filename)
{
    int fd = open(filename, O_RDONLY);
    if (fd < 0)
	quitter.fatal_error_with_errno("open %s", filename);
    struct stat st;
    if (fstat(fd, &st) < 0)
	quitter.fatal_error_with_errno("stat %s", filename);

    directory_entry *dep = volume->find(filename);
    if (!dep)
    {
	dep = volume->find("/");
	assert(dep);
	mode_t mode = S_IFREG + (0666 & ~get_umask());
	int err = dep->mknod(filename, mode, 0);
	if (err < 0)
	{
	    errno = -err;
	    quitter.fatal_error_with_errno
	    (
		"create %s:%s",
		volume->get_volume_name().c_str(),
		filename
	    );
	}
	dep = volume->find(filename);
	assert(dep);
    }

    int err = dep->open();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "open %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }

    err = dep->truncate(st.st_size);
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "truncate %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }

    off_t offset = 0;
    for (;;)
    {
	char buffer[1 << 14];
	ssize_t n = read(fd, buffer, sizeof(buffer));
	if (n < 0)
	    quitter.fatal_error_with_errno("read %s", filename);
	if (n == 0)
	    break;
	err = dep->write(offset, buffer, n);
	if (err < 0)
	{
	    errno = -err;
	    quitter.fatal_error_with_errno
    	    (
		"write %s:%s",
		volume->get_volume_name().c_str(),
		filename
	    );
	}
	offset += n;
    }
    err = dep->flush();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "flush %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
    err = dep->release();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
	(
	    "release %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
    close(fd);
}


static void
remove(directory *volume, const char *filename)
{
    directory_entry *dep = volume->find(filename);
    if (!dep)
    {
	errno = ENOENT;
	quitter.fatal_error_with_errno
	(
	    "remove %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
    int err = dep->unlink();
    if (err < 0)
    {
	errno = -err;
	quitter.fatal_error_with_errno
       	(
	    "unlink %s:%s",
	    volume->get_volume_name().c_str(),
	    filename
	);
    }
}


int
main(int argc, char **argv)
{
    progname_set(argv[0]);
    bool listing_flag = false;
    bool get_flag = false;
    bool put_flag = false;
    bool remove_flag = false;
    const char *disk_image_filename = 0;
    for (;;)
    {
	int c = getopt(argc, argv, "Df:glprV");
	if (c == EOF)
	    break;
	switch (c)
	{
	case 'D':
	    ++debug_level;
	    break;

	case 'f':
	    disk_image_filename = optarg;
	    break;

	case 'g':
	    get_flag = true;
	    break;

	case 'l':
	    listing_flag = true;
	    break;

	case 'p':
	    put_flag = true;
	    break;

	case 'r':
	    remove_flag = true;
	    break;

	case 'V':
	    version_print();
	    return 0;

	default:
	    usage();
	}
    }
    if (!disk_image_filename)
    {
	if (optind >= argc)
	    quitter.fatal_error("no disk image file name specified");
	// Ugly, but probably what the user meant.
	disk_image_filename = argv[optind++];
    }
    if (!listing_flag && !get_flag && !put_flag && !remove_flag)
	listing_flag = true;
    if (get_flag + put_flag + remove_flag > 1)
	usage();
    if (optind < argc && !get_flag && !put_flag && !remove_flag)
	usage();

    //
    // Open the volume, and make sure it has the right format.
    //
    bool read_only_flag = !put_flag && !remove_flag;
    directory *volume = directory::factory(disk_image_filename, read_only_flag);

    //
    // Chew on the volume, as requested.
    //
    while (optind < argc)
    {
	const char *filename = argv[optind++];
	if (put_flag)
	    put(volume, filename);
	if (get_flag)
	    get(volume, filename);
	if (remove_flag)
	    remove(volume, filename);
    }

    //
    // Show the contents of the volume.
    //
    if (listing_flag)
	volume->print_listing();

    //
    // Close down the volume.
    // This may do essential flush operations.
    //
    delete volume;
    volume = 0;

    //
    // Report success
    //
    return 0;
}
