//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#include <cassert>
#include <cerrno>
#include <cstdio>
#include <cstring>
#include <unistd.h>

#include <lib/debug.h>
#include <lib/directory.h>
#include <lib/directory/entry.h>
#include <lib/fuse.h>
#include <lib/hexdump.h>
#include <lib/progname.h>
#include <lib/quit.h>
#include <lib/rcstring/list.h>
#include <lib/sector_io/raw.h>
#include <lib/version.h>


static directory *volume;


static int
getattr_callback(const char *path, struct stat *stbuf)
{
    DEBUG(1, "getattr(path = \"%s\", stbuf = %p)", path, stbuf);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->getattr(stbuf);
}


static int
readlink_callback(const char *path, char *buf, size_t size)
{
    DEBUG(1, "readlink(path = \"%s\", buf = %p, size = %d)", path, buf,
	(int)size);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->readlink(buf, size);
}


static int
mknod_callback(const char *s, mode_t mode, dev_t dev)
{
    DEBUG(1, "mknod(path = \"%s\", mode = 0%o, dev = %d)", s, (int)mode,
	(int)dev);
    rcstring path(s);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (dep)
	return -EEXIST;
    dep = volume->find(path.dirname());
    if (!dep)
	return -ENOENT;
    return dep->mknod(path.basename(), mode, dev);
}


static int
mkdir_callback(const char *s, mode_t mode)
{
    DEBUG(1, "mkdir(s = \"%s\", mode = 0%o)", s, mode);
    rcstring path(s);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (dep)
	return -EEXIST;
    dep = volume->find(path.dirname());
    if (!dep)
	return -ENOENT;
    return dep->mkdir(path.basename(), mode);
}


static int
unlink_callback(const char *path)
{
    DEBUG(1, "unlink(path = \"%s\")", path);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->unlink();
}


static int
rmdir_callback(const char *path)
{
    DEBUG(1, "rmdir(path = \"%s\")", path);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->rmdir();
}


static int
symlink_callback(const char *lhs, const char *rhs)
{
    DEBUG(1, "symlink(lhs = \"%s\", rhs = \"%s\")", lhs, rhs);
    rcstring path(lhs);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (dep)
	return -EEXIST;
    dep = volume->find(path.dirname());
    if (!dep)
	return -ENOENT;
    return dep->symlink(path.basename(), rhs);
}


static int
rename_callback(const char *lhs, const char *rhs)
{
    DEBUG(1, "rename(lhs = \"%s\", rhs = \"%s\")", lhs, rhs);
    assert(volume);
    directory_entry *dep = volume->find(lhs);
    if (!dep)
	return -ENOENT;
    // FIXME: this isn't right, if moving between directories.
    return dep->rename(rhs);
}


static int
link_callback(const char *lhs, const char *rhs)
{
    DEBUG(1, "link(lhs = \"%s\", rhs = \"%s\")", lhs, rhs);
    rcstring path(lhs);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (dep)
	return -EEXIST;
    dep = volume->find(path.dirname());
    if (!dep)
	return -ENOENT;
    directory_entry *rhs_dep = volume->find(rhs);
    if (!rhs)
	return -ENOENT;
    return dep->link(path.basename(), rhs_dep);
}


static int
chmod_callback(const char *path, mode_t mode)
{
    DEBUG(1, "chmod(path = \"%s\", mode = 0%o)", path, mode);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->chmod(mode);
}


static int
chown_callback(const char *path, uid_t uid, gid_t gid)
{
    DEBUG(1, "chown(path = \"%s\", uid = %d, gid = %d)", path, uid, gid);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->chown(uid, gid);
}


static int
truncate_callback(const char *path, off_t size)
{
    DEBUG(1, "truncate(path = \"%s\", size = %ld)", path, (long)size);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->truncate(size);
}


static int
utime_callback(const char *path, struct utimbuf *buf)
{
    DEBUG(1, "utime(path = \"%s\", buf = %p)", path, buf);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->utime(buf);
}


static int
open_callback(const char *path, fuse_file_info *fi)
{
    DEBUG(1, "open(path = \"%s\", fi = %p)", path, fi);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    int err = dep->open();
    if (err < 0)
	return err;
    fi->fh = (unsigned long)dep;
    return 0;
}


static int
read_callback(const char *path, char *buf, size_t size, off_t offset,
    fuse_file_info *fi)
{
    DEBUG(1, "read(path = \"%s\", buf = %p, size = %ld, offset = %ld, fi = %p)",
	path, buf, (long)size, (long)offset, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    return dep->read(offset, buf, size);
}


static int
write_callback(const char *path, const char *buf, size_t size, off_t offset,
    fuse_file_info *fi)
{
    DEBUG(1, "write(path = \"%s\", buf = %p, size = %ld, offset = %ld, "
	"fi = %p)", path, buf, (long)size, (long)offset, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
     int err = dep->write(offset, buf, size);
     if (err < 0)
	 return err;
     return size;
}


static int
statfs_callback(const char *path, struct statfs *buf)
{
    DEBUG(1, "statfs(path = \"%s\", buf = %p)", path, buf);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->statfs(buf);
}


static int
flush_callback(const char *path, fuse_file_info *fi)
{
    DEBUG(1, "flush(path = \"%s\", fi = %p)", path, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    return dep->flush();
}


static int
release_callback(const char *path, fuse_file_info *fi)
{
    DEBUG(1, "release(path = \"%s\", fi = %p)", path, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    return dep->release();
}


static int
fsync_callback(const char *path, int arg, fuse_file_info *fi)
{
    DEBUG(1, "fsync(path = \"%s\", arg = %d, fi = %p)", path, arg, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    return dep->fsync(arg);
}


static int
setxattr_callback(const char *path, const char *name, const char *value,
    size_t size, int arg)
{
    DEBUG(1, "setxattr(path = \"%s\", name = \"%s\", value = %s, size = %ld, "
	"arg = %d)", path, name, rcstring(value, size).quote_c().c_str(),
	(long)size, arg);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->setxattr(name, value, size, arg);
}


static int
getxattr_callback(const char *path, const char *name, char *value, size_t size)
{
    DEBUG(1, "getxattr(path = \"%s\", name = \"%s\", value = %p, size = %ld)",
	path, name, value, (long)size);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->getxattr(name, value, size);
}


static int
listxattr_callback(const char *path, char *buf, size_t size)
{
    DEBUG(1, "listxattr(path = \"%s\", buf = %p, size = %ld)", path, buf,
	(long)size);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->listxattr(buf, size);
}


static int
removexattr_callback(const char *path, const char *name)
{
    DEBUG(1, "removexattr(path = \"%s\", name = \"%s\")", path, name);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    return dep->removexattr(name);
}


static int
opendir_callback(const char *path, fuse_file_info *fi)
{
    DEBUG(1, "opendir(path = \"%s\", fi = %p)", path, fi);
    assert(volume);
    directory_entry *dep = volume->find(path);
    if (!dep)
	return -ENOENT;
    fi->fh = (unsigned long)dep;
    return dep->opendir();
}


/**
  * Read directory
  *
  * This supersedes the old getdir() interface.  New applications
  * should use this.
  *
  * The filesystem may choose between two modes of operation:
  *
  * 1) The readdir implementation ignores the offset parameter,
  * and passes zero to the filler function's offset.  The filler
  * function will not return '1' (unless an error happens), so the
  * whole directory is read in a single readdir operation.  This
  * works just like the old getdir() method.
  *
  * 2) The readdir implementation keeps track of the offsets of the
  * directory entries.  It uses the offset parameter and always
  * passes non-zero offset to the filler function.  When the buffer
  * is full (or an error happens) the filler function will return
  * '1'.
  *
  * @returns
  *     zero on success, -errno on error.
  */
static int
readdir_callback(const char *path, void *buf, fuse_fill_dir_t filler,
    off_t offset, fuse_file_info *fi)
{
    DEBUG(1, "readdir(path = \"%s\", buf = %p, filler = %p, offset = %ld, "
	"fi = %p)", path, buf, filler, (long)offset, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    rcstring_list names;
    int err = dep->get_directory_entry_names(names);
    if (err < 0)
	return err;

    struct stat st;
    memset(&st, 0, sizeof(st));
    filler(buf, ".", &st, 0);
    filler(buf, "..", &st, 0);
    for (size_t j = 0; j < names.size(); ++j)
    {
	if (filler(buf, names[j].c_str(), &st, 0))
	    break;
    }
    return 0;
}


static int
releasedir_callback(const char *path, fuse_file_info *fi)
{
    DEBUG(1, "releasedir(path = \"%s\", fi = %p)", path, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    int result = dep->releasedir();
    fi->fh = 0;
    return result;
}


static int
fsyncdir_callback(const char *path, int arg, fuse_file_info *fi)
{
    DEBUG(1, "fsyncdir(path = \"%s\", arg = %d, fi = %p)", path, arg, fi);
    directory_entry *dep = (directory_entry *)fi->fh;
    if (!dep)
    {
	assert(volume);
	dep = volume->find(path);
    }
    if (!dep)
	return -ENOENT;
    return dep->fsyncdir(arg);
}


static fuse_operations ops =
{
    getattr_callback,
    readlink_callback,
    0, // getdir
    mknod_callback,
    mkdir_callback,
    unlink_callback,
    rmdir_callback,
    symlink_callback,
    rename_callback,
    link_callback,
    chmod_callback,
    chown_callback,
    truncate_callback,
    utime_callback,
    open_callback,
    read_callback,
    write_callback,
    statfs_callback,
    flush_callback,
    release_callback,
    fsync_callback,
    setxattr_callback,
    getxattr_callback,
    listxattr_callback,
    removexattr_callback,
    opendir_callback,
    readdir_callback,
    releasedir_callback,
    fsyncdir_callback,
    0, // init
    0, // destroy
};


static void
usage()
{
    const char *prog = progname_get();
    fprintf(stderr, "Usage: %s [ <option>... ] <volume> <dir>\n", prog);
    fprintf(stderr, "       %s -V\n", prog);
    quitter.exit(1);
}


int
main(int argc, char **argv)
{
    //
    // Parse the command line options.
    //
    progname_set(argv[0]);
    rcstring_list subset;
    rcstring_list mount_options;
    subset.push_back(progname_get());
    bool read_only_flag = false;
    for (;;)
    {
	int c = getopt(argc, argv, "Ddfho:rV");
	if (c < 0)
	    break;
	switch (c)
	{
	case 'D':
	    ++debug_level;
	    break;

	case 'd':
	    // debug
	    subset.push_back("-d");
	    break;

	case 'f':
	    // foreground
	    subset.push_back("-f");
	    break;

	case 'h':
	    // help - this is for compatibility with fusermount
	    version_print();
	    return 0;

	case 'o':
	    // mount option
	    mount_options.split(optarg, ",");
	    break;

	case 'r':
	    // read only
	    subset.push_back("-r");
	    read_only_flag = true;
	    break;

	case 'V':
	    version_print();
	    return 0;

	default:
	    usage();
	}
    }
    if (optind + 2 != argc)
	usage();
    const char *filename = argv[optind];
    const char *mount_point = argv[optind + 1];

    //
    // Look in the mount options to see if there is a "ro" option, it
    // means the same as the -r option.
    //
    if (mount_options.member("ro"))
    {
	mount_options.remove("ro");
	if (!read_only_flag)
	{
	    subset.push_back("-r");
	    read_only_flag = true;
	}
    }

    //
    // Look in the mount options to see if there is a umask=NNN option.
    // If not, insert one based on the current process' umask.
    //
    bool have_umask = false;
    for (size_t j = 0; j < mount_options.size(); ++j)
    {
	if (0 == memcmp(mount_options[j].c_str(), "umask=", 6))
	{
	    have_umask = true;
	    break;
	}
    }
    if (!have_umask)
    {
	mode_t um = umask(0);
	umask(um);
	rcstring s = rcstring::printf("umask=%05o", int(um));
	mount_options.push_back(s);
    }

    //
    // Add the options to the command line subset to be passed to fusermount
    //
    subset.push_back("-o" + mount_options.unsplit(","));

    //
    // Always run single threadded, that we we don't need locks.  It's
    // a pretty small and simple file system, mainly intended for
    // retrieving historical files, so performance is unlikely to matter
    // much.
    //
    subset.push_back("-s");

    //
    // Add the mount point as the last command line option
    // to the fusermount call.
    //
    subset.push_back(mount_point);

    //
    // Open the volume, and make sure it has the right format.
    //
    volume = directory::factory(filename, read_only_flag);

    //
    // Mount the file systems and start the fuse loop running to handle
    // requests from the kernel.
    //
    char **av = new char * [subset.size() + 1];
    for (size_t j = 0; j < subset.size(); ++j)
	av[j] = (char *)subset[j].c_str();
    av[subset.size()] = 0;
    fuse_main(subset.size(), av, &ops);

    //
    // Close down the volume.
    // This may do essential flush operations.
    //
    delete volume;
    volume = 0;

    //
    // Report success.
    //
    return 0;
}
