//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <utime.h>

#include <lib/quit.h>
#include <lib/rcstring.h>
#include <ucsdpsys_text/filter.h>


filter::~filter()
{
}


filter::filter()
{
}


void
filter::insitu(const rcstring &ifn)
{
    FILE *ifp = fopen(ifn.c_str(), "rb");
    if (!ifp)
	quitter.fatal_error_with_errno("open %s", ifn.quote_c().c_str());
    long name_max = fpathconf(fileno(ifp), _PC_NAME_MAX);
    if (name_max <= 0)
	name_max = 15;

    //
    // Get the access time and modified time, so we can restore them later.
    //
    struct stat st;
    if (fstat(fileno(ifp), &st) < 0)
	quitter.fatal_error_with_errno("stat %s", ifn.quote_c().c_str());

    //
    // Now we have to create a temporary file to write the translated
    // stuff to, before we rename it over the top of the original.
    //
    rcstring dir(ifn.dirname());
    rcstring base(ifn.basename());
    rcstring ofn;
    for (int num = 1; ; ++num)
    {
        rcstring numstr = rcstring::printf(".t%d", num);
	ofn =
	    rcstring::printf
	    (
		"%s/.%*s%s",
		dir.c_str(),
		(int)(name_max - 1 - numstr.size()),
		base.c_str(),
		numstr.c_str()
	    );
	if (access(ofn.c_str(), F_OK) < 0)
	    break;
    }

    FILE *ofp = fopen(ofn.c_str(), "wb");
    if (!ofp)
	quitter.fatal_error_with_errno("open %s", ifn.quote_c().c_str());

    translate(ifp, ifn, ofp, ofn);

    if (fclose(ifp))
	quitter.fatal_error_with_errno("close %s", ifn.quote_c().c_str());
    if (fclose(ofp))
	quitter.fatal_error_with_errno("close %s", ofn.quote_c().c_str());

    //
    // Now rename the new file over the top of the old file.
    //
    if (rename(ofn.c_str(), ifn.c_str()) < 0)
    {
	quitter.fatal_error_with_errno
        (
	    "mv %s %s",
	    ofn.quote_shell().c_str(),
	    ifn.quote_shell().c_str()
	);
    }

    //
    // Now restore the access time and modified time.
    //
    struct utimbuf amt;
    amt.actime = st.st_atime;
    amt.modtime = st.st_mtime;
    if (utime(ifn.c_str(), &amt) < 0)
	quitter.fatal_error_with_errno("utime %s", ofn.quote_c().c_str());
}


void
filter::run(int argc, char **argv)
{
    if (argc == 0)
    {
	translate(stdin, "stdin", stdout, "stdout");
    }
    else
    {
	for (int j = 0; j < argc; ++j)
	    insitu(argv[j]);
    }
}
