//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//
//	The special format of a textfile is document in the
//	ucsdpsys_text(1) man page.

#include <cctype>

#include <lib/quit.h>
#include <lib/rcstring.h>
#include <ucsdpsys_text/filter/decode.h>


filter_decode::~filter_decode()
{
}


filter_decode::filter_decode() :
    column(0),
    non_white(false)
{
}


void
filter_decode::process(const unsigned char *buffer, size_t size, FILE *ofp,
    const rcstring &ofn)
{
    while (size-- > 0)
    {
	unsigned char c = *buffer++;
	switch (c)
	{
	case '\0':
	    break;

	case '\r':
	case '\n':
	    putc('\n', ofp);
	    column = 0;
	    non_white = false;
	    break;

	case 16:
	    {
		if (size-- <= 0)
		{
                    // In theory, we will never see a (16,n) sequence
                    // span a 1KB boundary.  So we ignore the problem.
		    goto normal;
		}
		unsigned char c2 = *buffer++;
		if (c2 < 32)
		{
		    ++size;
		    --buffer;
		    goto normal;
		}
		c2 -= 32;
		if (non_white)
		{
                    // Insert spaces here.
		    // We only insert tabs at the start of the line.
		    for (unsigned x = 0; x < c2; ++x)
			putc(' ', ofp);
		}
		column += c2;
	    }
	    break;

	default:
	    normal:
	    if (!non_white)
	    {
		int ocol = 0;
		for (;;)
		{
		    if (ocol + 1 == column)
			break;
		    int ocol2 = (ocol + 8) & ~7;
		    if (ocol2 > column)
			break;
		    putc('\t', ofp);
		    ocol = ocol2;
		}
		while (ocol < column)
		{
		    putc(' ', ofp);
		    ++ocol;
		}
		non_white = true;
	    }
	    putc(c, ofp);
	    ++column;
	    break;
	}
    }
    if (ferror(ofp))
	quitter.fatal_error_with_errno("write %s", ofn.quote_c().c_str());
}


static inline bool
is_text_character(unsigned char c)
{
    return (isprint(c) || isspace(c) || c == 16);
}


static bool
is_text_buffer(unsigned char *buffer, size_t size)
{
    for (size_t j = 0; j < 16 && j < size; ++j)
	if (!is_text_character(buffer[j]))
	    return false;
    return true;
}


void
filter_decode::translate(FILE *ifp, const rcstring &ifn, FILE *ofp,
    const rcstring &ofn)
{
    //
    // Read the first 1KB into a buffer.  It could be the editor
    // settings block, and it could be text.  Sniff and see.
    //
    unsigned char buffer[1024];
    size_t n = fread(buffer, 1, sizeof(buffer), ifp);
    if (n < sizeof(buffer))
    {
	//
	// It can't a an editor settings structure or it would be
	// exactly 1KB.
	//
	if (ferror(ifp))
	    quitter.fatal_error_with_errno("read %s", ifn.quote_c().c_str());
	short_read:
	if (n)
	    process(buffer, n, ofp, ofn);
	if (column != 0 && non_white)
	    putc('\n', ofp);
	return;
    }
    if (is_text_buffer(buffer, sizeof(buffer)))
	process(buffer, sizeof(buffer), ofp, ofn);

    //
    // Process the rest of the file.
    //
    for (;;)
    {
	n = fread(buffer, 1, sizeof(buffer), ifp);
	if (n < sizeof(buffer))
	    goto short_read;
	process(buffer, sizeof(buffer), ofp, ofn);
    }
    if (column != 0 && non_white)
	putc('\n', ofp);
    if (fflush(ofp))
	quitter.fatal_error_with_errno("write %s", ofn.quote_c().c_str());
}
