//
// UCSD p-System filesystem in user space
// Copyright (C) 2006 Peter Miller
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at
// you option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston MA 02111-1307,
// USA
//

#include <cassert>

#include <lib/quit.h>
#include <lib/rcstring/accumulator.h>
#include <ucsdpsys_text/filter/encode.h>


filter_encode::~filter_encode()
{
}


filter_encode::filter_encode()
{
}


static bool
get_one_line(FILE *ifp, const rcstring &ifn, rcstring &line)
{
    static rcstring_accumulator ac;
    ac.clear();
    int column = 0;
    bool non_white = false;
    for (;;)
    {
	int c = getc(ifp);
	switch (c)
	{
	case EOF:
	    if (ferror(ifp))
	    {
		quitter.fatal_error_with_errno
		(
		    "read %s",
		    ifn.quote_c().c_str()
		);
	    }
	    if (!ac.empty())
	    {
		ac.push_back('\r');
		line = ac.mkstr();
		return true;
	    }
	    return false;

	case ' ':
	    if (non_white)
		ac.push_back(' ');
	    ++column;
	    break;

	case '\t':
	    if (non_white)
		ac.push_back('\t');
	    column = (column + 8) & ~7;
	    break;

	// case 16:
	case '\0':
	    // Ignore
	    break;

	case '\n':
	case '\r':
	    ac.push_back('\r');
	    line = ac.mkstr();
	    return true;

	default:
	    if (!non_white)
	    {
		int ocol = 0;
		while (ocol + 223 <= column)
		{
		    ac.push_back((char)16);
		    ac.push_back((char)255);
		    ocol += 223;
		}
		switch (column - ocol)
		{
		case 0:
		    break;

		case 2:
		    ac.push_back(' ');
		    // Fall through...

		case 1:
		    ac.push_back(' ');
		    break;

		default:
		    ac.push_back((char)16);
		    ac.push_back((char)(column - ocol + 32));
		    break;
		}
		non_white = true;
	    }
	    ac.push_back((char)c);
	    break;
	}
    }
}


void
filter_encode::translate(FILE *ifp, const rcstring &ifn, FILE *ofp,
    const rcstring &ofn)
{
    // The special format of a textfile is as follows:
    //
    // [] There are two blocks (1024 bytes) of header information at the
    //    beginning of the file.  This information is used by the Pascal
    //    Editor.  The Pascal system creates the header page wheIl a
    //    user program opens a textfile.  The header page is transferred
    //    only during disk-to-disk transfers; transfers to character
    //    devices, such as the eonsole or printer, alwal ys omit the
    //    header page.
    //
    // [] The rest of the file consists of two-block pages.  Each
    //    page contains lines of text, separated from each other by
    //    RETURN characters (ASCII 13).  No line ever crosses a page
    //    boundary; thus a page contains only whole lines.  After the
    //    last line on a page, the remainder of the page is filled with
    //    NUL characters (ASCII OO).  READ and READLN skip the NUL
    //    characters, and WRITE and WRITELN provide them automatically.
    //    Thus this page formatting is normally invisible to a Pascal
    //    program.
    //
    // [] A sequence of leading spaces in a line may be compressed to a
    //    DLE-blank code.  This code consists of a DLE control character
    //    (ASCII 16) followed by one byte containing the number of
    //    spaces to indent plus 32 (decimal).  Using this code saves a
    //    considerable amount of space in files where indentation occurs
    //    frequently.  The Editor is the main creator of DLE-blank
    //    codes; it usually outputs a DLE-blank eode where a sequence
    //    of spaces occurs at the beginning of a line.  However, the
    //    DLE-blank code is optional; some lines may have it, and others
    //    may have space characters instead.  Also, a line with no
    //    indentation may or may not be preceded by a DLE character and
    //    an indent code value of 32 (meaning O indentation).
    //
    for (int j = 0; j < 1024; ++j)
    {
	putc('\0', ofp);
    }
    long address = 0;
    for (;;)
    {
	assert(address < 1024);
	rcstring line;
	if (!get_one_line(ifp, ifn, line))
	    break;
	if (address != 0 && address + line.size() > 1024)
	{
	    for (;;)
	    {
		putc(0, ofp);
		++address;
		if ((address & 1023) == 0)
		    break;
	    }
	}
	fwrite(line.c_str(), 1, line.size(), ofp);
	address = (address + line.size()) & 1023;
	if (ferror(ofp))
	    goto barf;
    }

    // Should the last block be padded with NULs?
    if (address)
    {
	for (;;)
	{
	    putc(0, ofp);
	    ++address;
	    if ((address & 1023) == 0)
		break;
	}
    }

    if (fflush(ofp))
    {
	barf:
	quitter.fatal_error_with_errno("write %s", ofn.quote_c().c_str());
    }
}
